#include <Wire.h>
#include <lcdgfx.h>
#include <nano_engine_v2.h>
#include <nano_gfx_types.h>
//#include "heltec.h"

#include "micro_keyboard.h"

DisplaySSD1306_128x64_I2C display(RST_OLED, {-1, 0x3c, SCL_OLED, SDA_OLED, 700000});
typedef NanoEngine1<DisplaySSD1306_128x64_I2C> MyNanoEngine;
MyNanoEngine engine(display);
MicroKeyboard kb;
uint8_t *kbState;
SemaphoreHandle_t ledControl;

const PROGMEM uint8_t heartImage[8] = {
  0b00001110,
  0b00011111,
  0b00111111,
  0b01111110,
  0b01111110,
  0b00111101,
  0b00011001,
  0b00001110
};

const PROGMEM uint8_t shiftImage[8] = {
  0b00000000,
  0b00001000,
  0b00001100,
  0b11111110,
  0b11111110,
  0b00001100,
  0b00001000,
  0b00000000
};

NanoFixedSprite<MyNanoEngine::TilerT> heart({60, 28}, {8, 8}, heartImage);
NanoFixedSprite<MyNanoEngine::TilerT> shiftSprite({128, 0}, {8, 8}, shiftImage);

class TextThing: public NanoObject<MyNanoEngine::TilerT> {
public:
  TextThing(const NanoPoint &pos, String initial_text) : NanoObject<MyNanoEngine::TilerT>(pos, {0, 8}) {
    text = initial_text;
    setSize({text.length() * 6, 8});
  }
  
  void draw() override {
    this->getTiler().getCanvas().printFixed(this->getPosition().x, this->getPosition().y, text.c_str(), STYLE_NORMAL);
  }

  void updateFromKeyPress(KeyPress *kp) {
    if (kp->key == MKB_BKSP) {
      resize({text.length() * 6, 8});
      text.remove(text.length() - 1);
      return;
    } else if (kp->key == MKB_ENTER) {
      return;
    } else if (kp->key >= 0x80) {
      return;
    } else if (kp->modifierState & (MKB_MOD_CTRL | MKB_MOD_ALT | MKB_MOD_FN | MKB_MOD_SYM)) {
      return;
    } else if (kp->modifierState & MKB_MOD_SHIFT) {
      text += (char) kp->key;
    } else {
      text += (char) (kp->key | 0x20);
    }

    resize({text.length() * 6, 8});
  }

private:
  String text;
};

TextThing tt({1, 55}, "");

bool drawAll() {
  engine.getCanvas().clear();
  engine.getCanvas().drawRect(0, 0, 127, 63);
  engine.getCanvas().drawLine(0, 0, 127, 63);
  engine.getCanvas().drawLine(127, 0, 0, 63);

  uint8_t modifierState = kb.getModifierState();
  char buf[80];
  sprintf(buf, "%02x %02x %02x %02x %02x - %02x", kbState[0], kbState[1], kbState[2], kbState[3], kbState[4], modifierState);
  if (modifierState & MKB_MOD_SHIFT) {
    shiftSprite.moveTo({120, 0});
  } else {
    shiftSprite.moveTo({128, 0});    
  }
  engine.getCanvas().printFixed(0, 0, buf, STYLE_NORMAL);
  engine.refresh({ {0, 0}, {128, 8} });
  return true;
}

void blinkLed(void *parameter) {
  bool state = true;
  while (1) {
    if (state) {
      xSemaphoreTake(ledControl, portMAX_DELAY);
      xSemaphoreGive(ledControl);
    }
    digitalWrite(LED, state);
    state = !state;
    delay(1000);
  }
}

void setup() {
  pinMode(Vext,OUTPUT);
  digitalWrite(Vext, LOW);
  
  //Heltec.begin(false /*DisplayEnable Enable*/, true /*LoRa Disable*/, true /*Serial Enable*/);
  display.begin();
  //Wire.begin(SDA_OLED, SCL_OLED);
  Wire.setClock(700000);
  display.getInterface().flipHorizontal(1);
  display.getInterface().flipVertical(1);

  engine.setFrameRate(60);
  engine.getCanvas().setFixedFont(ssd1306xled_font6x8);
  engine.drawCallback(drawAll);
  engine.insert(heart);
  engine.insert(tt);
  engine.insert(shiftSprite);
  engine.refresh();
  
  pinMode(LED, OUTPUT);

  Serial.begin(115200);

  ledControl = xSemaphoreCreateBinary();
  xTaskCreate(blinkLed, "Blink LED", configMINIMAL_STACK_SIZE, NULL, 1, NULL);
}

void loop() {
  if (engine.nextFrame()) {
    kb.scan();
    kbState = kb.getState();
  
    if (kbState[0] & 0x10) {
      heart.moveBy({-1, 0});
    }
    if (kbState[0] & 0x80) {
      heart.moveBy({1, 0});
    }
    if (kbState[0] & 0x20) {
      heart.moveBy({0, 1});
    }
    if (kbState[0] & 0x40) {
      heart.moveBy({0, -1});
    }
    if (heart.x() < 0) {
      heart.moveBy({120, 0});
    } else if (heart.x() > 120) {
      heart.moveBy({-120, 0});
    }
    if (heart.y() < 0) {
      heart.moveBy({0, 56});
    } else if (heart.y() > 56) {
      heart.moveBy({0, -56});
    }

    KeyPress kp;
    while (kb.getKeyPress(&kp)) {
      tt.updateFromKeyPress(&kp);
      if (kp.modifierState & MKB_MOD_FN) {
        if (kp.key == 'G') {
          xSemaphoreGive(ledControl);
        } else if (kp.key == 'T') {
          xSemaphoreTake(ledControl, 0);
        }
      }
    }

    engine.update();
    engine.display();

    /*
    char buf[255];
    sprintf(buf, "draw %d\n", engine.getCpuLoad());
    Serial.write(buf);
    */
  }
}
