#include <stdint.h>
#include <Arduino.h>

#include "micro_keyboard.h"

#define CN0 12
#define CN1 13
#define CN2 33
#define R1 36
#define R2 37
#define R3 38
#define R4 39
#define R5 32

MicroKeyboard::MicroKeyboard() {
  memset(state, 0, sizeof(state));
  modifierState = 0;
  multiModMode = false;

  pinMode(CN0, OUTPUT);
  pinMode(CN1, OUTPUT);
  pinMode(CN2, OUTPUT);
  pinMode(R1, INPUT);
  pinMode(R2, INPUT);
  pinMode(R3, INPUT);
  pinMode(R4, INPUT);
  pinMode(R5, INPUT);
}

bool MicroKeyboard::keyEventInWindow(uint8_t key, bool pressed, unsigned long t_lower_bound) {
  for (int i = 0; i < keyEventBuffer.numElements(); i++) {
    auto k = keyEventBuffer.peek(i);
    if (k->key == key && k->pressed == pressed && k->ts >= t_lower_bound) {
      return true;
    }
  }
  return false;
}

void MicroKeyboard::scan() {
  memcpy(this->previousState, this->state, ROWS);
  memset(this->state, 0, ROWS);
  
  for (unsigned char col = 0; col < 8; col++) {
    digitalWrite(CN0, bitRead(col, 0));
    digitalWrite(CN1, bitRead(col, 1));
    digitalWrite(CN2, bitRead(col, 2));
    delayMicroseconds(10);
    this->state[0] |= !digitalRead(R1) << col;
    this->state[1] |= !digitalRead(R2) << col;
    this->state[2] |= !digitalRead(R3) << col;
    this->state[3] |= !digitalRead(R4) << col;
    this->state[4] |= !digitalRead(R5) << col;
  }
  auto t0 = millis();
  auto t_window = t0 - 50;
  digitalWrite(CN0, 0);
  digitalWrite(CN1, 0);
  digitalWrite(CN2, 0);

  if (!keyPressBuffer.isFull()) {
    for (unsigned char row = 0; row < 5; row++) {
      uint8_t pressed = this->state[row] & ~this->previousState[row];
      if (pressed) {
        for (unsigned char col = 0; col < 8; col++) {
          auto key = keymap[row][col];
          if (keyEventInWindow(key, true, t_window)) {
            // key has already been pressed lately; ignore it
            continue;
          }
          if (pressed & (1 << col)) {
            keyEventBuffer.add(KeyEvent { key, true, t0 }, true);
            
            if (key >= 0x90 && key <= 0x9F) {
              switch (key) {
                case MKB_SHIFT:
                  modifierState ^= MKB_MOD_SHIFT;
                  break;
                case MKB_LCTRL:
                case MKB_RCTRL:
                  modifierState ^= MKB_MOD_CTRL;
                  break;
                case MKB_ALT:
                  modifierState ^= MKB_MOD_ALT;
                  break;
                case MKB_SYM:
                  modifierState ^= MKB_MOD_SYM;
                  break;
                case MKB_FN:
                  modifierState ^= MKB_MOD_FN;
                  break;
              }
            }
          }
        }
      }
      uint8_t released = ~this->state[row] & this->previousState[row];
      if (released) {
        for (unsigned char col = 0; col < 8; col++) {
          auto key = keymap[row][col];
          if (keyEventInWindow(key, false, t_window)) {
            // key has already been released lately; ignore it
            continue;
          }
          if (released & (1 << col)) {
            keyEventBuffer.add(KeyEvent { key, false, t0 }, true);
            
            if (key >= 0x90 && key <= 0x9F) {
              if (multiModMode) {
                switch (key) {
                  case MKB_SHIFT:
                    modifierState &= ~MKB_MOD_SHIFT;
                    break;
                  case MKB_LCTRL:
                  case MKB_RCTRL:
                    modifierState &= ~MKB_MOD_CTRL;
                    break;
                  case MKB_ALT:
                    modifierState &= ~MKB_MOD_ALT;
                    break;
                  case MKB_SYM:
                    modifierState &= ~MKB_MOD_SYM;
                    break;
                  case MKB_FN:
                    modifierState &= ~MKB_MOD_FN;
                    break;
                }
                multiModMode = false;
              }
            } else {
              keyPressBuffer.add(KeyPress { key, modifierState });
              if (!(this->state[4] & 0b00111011 || this->state[2] & 0b00000001)) {
                // if no modifier keys are currently pressed, cancel modiferState
                modifierState = 0;
              } else {
                // well I guess we're holding it down. Cancel it when released.
                multiModMode = true;
              }
            }
          }
        }
      }
    }
  }
}

uint8_t *MicroKeyboard::getState() {
  return this->state;
}

uint8_t MicroKeyboard::getModifierState() {
  return this->modifierState;
}

bool MicroKeyboard::getKeyPress(KeyPress *kp) {
  return keyPressBuffer.pull(kp);
}
